#include <Drivers.h>
#include <KernelExport.h>
#include <OS.h>
#include <stdlib.h>
#include <sys/param.h>
#include <sys/types.h>
#include <unistd.h>

/*

	/dev/config/sysinfo driver for BeOS
	(c) 2002, mmu_man, revol@free.fr
	some kernel infos...
	usage:
		make driverinstall
		rescan sysinfo
		cat /dev/misc/sysinfo
	licenced under MIT licence
	NOT SMP-safe, NOT threadsafe (well, not sure)
*/

/*
*/

// this is really uint64 !!
extern uint64 max_swap_size;
extern uint64 swap_size;
extern uint64 used_swap_size;

extern uint32 nareas;
extern uint32 nsems;
extern uint32 nports;
extern uint32 nteams;
extern uint32 nthreads;

extern uint32 spur_int_cnt;

extern char *copyright;

extern uint32 memsize;

// locals to kernel_intel :-(
//extern uint32 secs;
//extern bigtime_t usecs;

extern uint32 boot_time;
extern uint64 time_base_to_usec;
extern uint32 timer_factor_shift;
extern uint32 timer_factor;

extern uint64 system_time_base;
extern uint64 apic_time_base_to_usec;
extern uint32 apic_timer_clknum_hz;

extern uint32 cv_factor;

extern uint32 kstk_size;
extern thread_id picasso_thid;


/* whatever it is, just use it to find the local ones */
extern char cpupreemption[100];

uint32 *p_secs;
bigtime_t *p_usecs;


/*
objdump --all-headers /system/kernel_intel >kernel.syms


address          section size    name
0008979c l     O .data  00000004 secs
000897a0 l     O .data  00000004 timezone_offset
000897a4 l     O .data  00000004 dst_observed
000897a8 l     O .data  00000001 rtc_stores_gmt
000897ac l     O .data  00000004 rtc_lock
000897b0 l     O .data  00000018 leap_month_table
000897c8 l     O .data  00000018 month_table
00067770 l     F .text  000000bd timeofday_to_secs
00067830 l     F .text  00000118 secs_to_timeofday
000897e0 l     O .data  00000001 uip_tested.39
000897e1 l     O .data  00000001 uip_always_on.40
00067948 l     F .text  0000043b readrtc_time
000a0324 l     O .bss   00000004 intr_en
00067d84 l     F .text  0000066e writertc_time
0009ff00 l     O .bss   00000008 usecs
0009ff08 l     O .bss   00000004 boot_time
0009ff20 l     O .bss   00000401 tzfilename

all those are local symbols :-(

000a6b00 g     O .bss   00000100 cpupreemption

000a6fdc g     O .bss   00000004 nareas

0004879c g     F .text  00000060 utime

000a9220 g     O .bss   00000008 time_base_to_usec

000889f0 g     O .data  00000004 spur_int_cnt

00087820 g     O .data  00000004 copyright

000a6fa0 g     O .bss   00000004 nthreads
000a707c g     O .bss   00000004 memsize

000a9228 g     O .bss   00000004 timer_factor_shift

000867a8 g     O .data  00000001 hlt_enabled

00086688 g     O .data  00000004 num_ints

000a922c g     O .bss   00000004 timer_factor

000a682c g     O .bss   00000004 maxmsgs
000a68a0 g     O .bss   00000004 nsems
0008985c g     O .data  00000004 pci_maxdevices
000a6fb4 g     O .bss   00000004 kstk_size

000867b4 g     O .data  00000004 picasso_thid

000896a0 g     O .data  00000004 cv_factor
000867ac g     O .data  00000004 nctx
000896a4 g     O .data  00000008 system_time_base
000a6830 g     O .bss   00000004 nports
000a6fc8 g     O .bss   00000004 nteams


*/

const char * device_names[]={"misc/sysinfo", NULL};
extern device_hooks sysinfo_hooks;

int32 api_version = B_CUR_DRIVER_API_VERSION;

#define THE_BUFF_SIZE 1024

typedef struct {
	char buff[THE_BUFF_SIZE];
	int len;
} cookie_t;


static void setup_pointers(void)
{
	p_secs = (uint32 *)(cpupreemption + 0x0008979c - 0x000a6b00);
	p_usecs = (bigtime_t *)(cpupreemption + 0x0009ff00 - 0x000a6b00);
}

status_t init_hardware(void) {
dprintf("sysinfo:init_hardware()\n");
	return B_OK;
}

status_t init_driver(void) {
dprintf("sysinfo:init_driver()\n");
	setup_pointers();
	return B_OK;
}

void uninit_driver(void) {
dprintf("sysinfo:uninit_driver()\n");
}

const char **publish_devices() {
	return device_names;
}

device_hooks *find_device(const char *name) {
	(void)name;
	return &sysinfo_hooks;
}

status_t sysinfo_open(const char *name, uint32 flags, cookie_t **cookie) {
	char *ptr;
	(void)name; (void)flags;
	*cookie = (void*)malloc(sizeof(cookie_t));
	if (*cookie == NULL) {
		dprintf("sysinfo_open : error allocating cookie\n");
		goto err0;
	}
	memset(*cookie, 0, sizeof(cookie_t));
	ptr = (*cookie)->buff;
	sprintf(ptr, "max_swap_size\t%ld\nswap_size\t%ld\nused_swap_size\t%ld\n", (long)max_swap_size, (long)swap_size, (long)used_swap_size);
	ptr += strlen(ptr);
	sprintf(ptr, "nareas\t%ld\nnsems\t%d\nnports\t%d\nnteams\t%d\nnthreads\t%d\n", nareas, nsems, nports, nteams, nthreads);
//	ptr += strlen(ptr);
//	sprintf(ptr, "copyright\t%s\n", copyright); /* 53 */
	ptr += strlen(ptr);
	sprintf(ptr, "spur_int_cnt\t%ld\nmemsize\t%ld\n", spur_int_cnt, memsize);
//	ptr += strlen(ptr);
//	sprintf(ptr, "kstk_size\t%ld\npicasso_thid\t%ld\n", kstk_size, picasso_thid);
	ptr += strlen(ptr);
	sprintf(ptr, "secs\t%ld\nusecs\t0x%08lx%08lx\n", *p_secs, (long)((*p_usecs) >> 32), (long)(*p_usecs));
	ptr += strlen(ptr);
	sprintf(ptr, "time_base_to_usec\t0x%08lx%08lx\n", (time_base_to_usec >> 32), time_base_to_usec);
	ptr += strlen(ptr);
	sprintf(ptr, "timer_factor_shift\t%ld\ntimer_factor\t%ld\n", timer_factor_shift, timer_factor);
/*
	ptr += strlen(ptr);
	sprintf(ptr, "system_time_base\t0x%08lx%08lx\n", (system_time_base >> 32), system_time_base);
	ptr += strlen(ptr);
	sprintf(ptr, "apic_time_base_to_usec\t0x%08lx%08lx\n", (apic_time_base_to_usec >> 32), apic_time_base_to_usec);
	ptr += strlen(ptr);
	sprintf(ptr, "apic_timer_clknum_hz\t%ld\n", apic_timer_clknum_hz);
*/
	ptr += strlen(ptr);
	sprintf(ptr, "cv_factor\t%ld\n", cv_factor);
//	ptr += strlen(ptr);
//	sprintf(ptr, "secs\t\t%ld\nusecs\t0x%08lx%08lx\n", secs, (usecs >> 32), usecs);

/*

extern uint64 time_base_to_usec;
extern uint32 timer_factor_shift;
extern uint32 timer_factor;


extern uint32 secs;
extern bigtime_t usecs
extern uint32 boot_time;
extern uint64 time_base_to_usec;
extern uint32 timer_factor_shift;
extern uint32 timer_factor;

extern uint32 kstk_size;
extern thread_id picasso_thid;
*/
	(*cookie)->len = strlen((*cookie)->buff);
	return B_OK;
	
err1:
	dprintf("sysinfo_open : cleanup : will free cookie\n");
	free(*cookie);
	*cookie = NULL;
err0:
	return B_ERROR;
}

status_t sysinfo_close(void *cookie) {
	(void)cookie;
	return B_OK;
}

status_t sysinfo_free(cookie_t *cookie) {
	free(cookie);
	return B_OK;
}

status_t sysinfo_ioctl(cookie_t *cookie, uint32 op, void *data, size_t len) {
	(void)cookie; (void)op; (void)data; (void)len;
	return B_ERROR;
}

status_t sysinfo_read(cookie_t *cookie, off_t position, void *data, size_t *numbytes) {
	int numread = 0;

	if (position > cookie->len)
		goto ERROR_EOF;
	while (position < cookie->len && position < *numbytes) {
		*(char *)data++ = cookie->buff[position++];
		numread++;
	}
	*numbytes = numread;
	return B_OK;

ERROR_EOF:
	*numbytes = 0;
	return B_OK;
}

status_t sysinfo_write(void *cookie, off_t position, const void *data, size_t *numbytes) {
	bool ena;
	(void)cookie; (void)position; (void)data;
//dprintf("sysinfo_write(, %ld, , %ld)\n", (long)position, (long)*numbytes);
ERROR_EOF:
	*numbytes = 0;
	return B_ERROR;
}

device_hooks sysinfo_hooks={
	(device_open_hook)sysinfo_open,
	sysinfo_close,
	(device_free_hook)sysinfo_free,
	(device_control_hook)sysinfo_ioctl,
	(device_read_hook)sysinfo_read,
	sysinfo_write,
	NULL,
	NULL,
	NULL,
	NULL
};
